/*
    This file is part of the FElt finite element analysis package.
    Copyright (C) 1993-1995 Jason I. Gobat and Darren C. Atkinson

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/************************************************************************
 * File:	generator.c						*
 *									*
 * Description:	This file contains the public and private function	*
 *		definitions which relate to the generator instance.	*
 ************************************************************************/

# include <stdio.h>
# include <string.h>
# include <unistd.h>
# include "error.h"
# include "generator.h"
# include "definition.h"

extern char *strdup ( );


# define streq(a,b)	!strcmp(a,b)
# define strneq(a,b)	strcmp(a,b)

# ifndef LIBDIR
# define LIBDIR "/usr/local/lib/felt"
# endif

# ifndef CPP
# define CPP "/lib/cpp"
# endif


Generator generator;

static char *cpp;
static char  cpp_command [2048];

# ifndef DOS
/************************************************************************
 * Function:	defnlookup						*
 *									*
 * Description:	Looks up a definition by name.				*
 ************************************************************************/

Definition defnlookup (name)
    char *name;
{
    Definition definition;

/*
    char *ptr;
    ptr = name;
    while (*ptr != ' ')
        ptr ++;

    *ptr = 0;
*/
    if (!(definition = LookupDefinition (name)))
        error ("%s elements have no definition", name);

    return definition;
}
# endif

/************************************************************************
 * Function:	ReadCorduroyFile					*
 *									*
 * Description:	Reads a corduroy file using the preprocessor if		*
 *		desired.  A filename of "-" indicates standard input	*
 *		(can only be used initially) and a NULL filename	*
 *		indicates no file (an empty generation instance is	*
 *		created).						*
 ************************************************************************/

int ReadCorduroyFile (filename)
    char *filename;
{
    char *plural;
    FILE *input;


    /* Open the file and send it through the preprocessor. */

    if (filename) {
	if (strneq (filename, "-")) {
	    if (access (filename, R_OK)) {
		error ("Unable to open %s", filename);
		return 1;
	    }

	    if (!freopen (filename, "r", stdin)) {
		error ("Unable to open %s", filename);
		return 1;
	    }
	}

# ifdef DOS
        input = stdin;
# else
	if (cpp) {
	    if (!(input = popen (cpp_command, "r"))) {
		error ("Unable to execute preprocessor");
		return 1;
	    }
	} else
	    input = stdin;
# endif

    } else
	input = NULL;


    /* Initialize the generator instance. */

    generator.start_node    = 1;
    generator.start_element = 1;
    generator.material      = NULL;
    generator.constraint    = NULL;
    generator.num_errors    = 0;
    generator.line	    = 1;
    generator.lines	    = NULL;
    generator.grids	    = NULL;
    generator.quadgrids	    = NULL;
    generator.trimeshes	    = NULL;
    generator.num_lines	    = 0;
    generator.num_grids	    = 0;
    generator.num_quadgrids = 0;
    generator.num_trimeshes = 0;

    if (filename)
	generator.filename = strdup (streq(filename, "-") ? "stdin" : filename);
    else
	generator.filename = strdup ("");


    /* Parse the input. */

    if (filename) {
	init_lexer (input);
	yyparse ( );
	generator.line = 0;

# ifdef DOS
        fclose (input);
# else
	if (cpp)
	    pclose (input);
	else
	    fclose (input);
# endif

	/* Report any errors. */

	if (generator.num_errors) {
	    plural = generator.num_errors != 1 ? "errors" : "error";
	    error ("%u %s found in input", generator.num_errors, plural);
	    return generator.num_errors;
	}
    }

    return 0;
}

# ifndef DOS
/************************************************************************
 * Function:	ParseCppOptions						*
 *									*
 * Description:	Parses and removes the preprocesor options from the	*
 *		command line arguments.					*
 ************************************************************************/

int ParseCppOptions (argc, argv)
    int  *argc;
    char *argv [ ];
{
    int   i;
    int   j;
    char *arg;
    char  cpp_args [2048];


    j = 1;
    cpp = CPP;

    for (i = 1; i < *argc; i ++)
	if ((arg = argv [i]) [0] != '-') {
	    argv [j ++] = arg;
	} else if (streq (arg, "-nocpp")) {
	    cpp = NULL;
	} else if (streq (arg, "-cpp")) {
	    if (++ i == *argc)
		return 1;
	    cpp = argv [i];
	} else if (arg [1] == 'D' || arg [1] == 'U' || arg [1] == 'I') {
	    strcat (cpp_args, " '");
	    strcat (cpp_args, arg);
	    strcat (cpp_args, "'");
	} else
	    argv [j ++] = arg;

    if (cpp)
	sprintf (cpp_command, "%s -I%s %s", cpp, LIBDIR, cpp_args);

    argv [*argc = j] = NULL;
    return 0;
}
# endif
